"""Data models for source code management functions."""

import re
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, MutableMapping, Optional, Protocol

from bumpversion.ui import get_indented_logger
from bumpversion.utils import Pathlike, extract_regex_flags

logger = get_indented_logger(__name__)

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.config import Config


@dataclass
class SCMConfig:
    """Configuration for source code management functions."""

    tag: bool
    sign_tags: bool
    tag_name: str
    allow_dirty: bool
    commit: bool
    message: str
    parse_pattern: str
    tag_message: Optional[str] = None
    commit_args: Optional[str] = None
    moveable_tags: List[str] = field(default_factory=list)

    def get_version_from_tag(self, tag: str) -> Optional[str]:
        """Return the version from a tag."""
        version_pattern = self.parse_pattern.replace("\\\\", "\\")
        version_pattern, regex_flags = extract_regex_flags(version_pattern)
        parts = self.tag_name.split("{new_version}", maxsplit=1)
        prefix = parts[0]
        # suffix = parts[1]
        rep = f"{regex_flags}{re.escape(prefix)}(?P<current_version>{version_pattern})"
        tag_regex = re.compile(rep)
        return match["current_version"] if (match := tag_regex.search(tag)) else None

    @classmethod
    def from_config(cls, config: "Config") -> "SCMConfig":
        """Return a SCMConfig from a Config object."""
        return cls(
            tag=config.tag,
            tag_name=config.tag_name,
            sign_tags=config.sign_tags,
            allow_dirty=config.allow_dirty,
            commit=config.commit,
            message=config.message,
            parse_pattern=config.parse,
            tag_message=config.tag_message,
            commit_args=config.commit_args,
            moveable_tags=config.moveable_tags,
        )


@dataclass
class LatestTagInfo:
    """Information about the latest tag."""

    commit_sha: Optional[str] = None
    distance_to_latest_tag: int = 0
    current_version: Optional[str] = None
    current_tag: Optional[str] = None
    branch_name: Optional[str] = None
    short_branch_name: Optional[str] = None
    repository_root: Optional[Path] = None
    dirty: Optional[bool] = None


class SCMTool(Protocol):
    """Protocol for source code management tools."""

    def __init__(self, config: SCMConfig): ...

    def is_available(self) -> bool:
        """Return whether the SCM tool is available."""
        ...

    def latest_tag_info(self) -> LatestTagInfo:
        """Return the latest tag information."""
        ...

    def add_path(self, path: Pathlike) -> None:
        """Add a path to the pending commit."""
        ...

    def get_all_tags(self) -> List[str]:
        """Return all tags in the SCM."""
        ...

    def commit_and_tag(self, files: List[Pathlike], context: MutableMapping, dry_run: bool = False) -> None:
        """Commit and tag files to the repository using the configuration."""
        ...

    def assert_nondirty(self) -> None:
        """
        Asserts that the repository is not dirty.

        Raises:
            DirtyWorkingDirectoryError: If the repository is not clean.
        """
        ...


class DefaultSCMTool:
    """Default implementation of the SCMTool protocol."""

    def __init__(self, config: SCMConfig):
        self.config = config

    def __repr__(self) -> str:
        """Return a string representation of the SCMTool."""
        return self.__str__()

    def __str__(self) -> str:
        """A string representation of the object."""
        return "None"

    def is_available(self) -> bool:
        """Return whether the SCM tool is available."""
        return True

    def latest_tag_info(self) -> LatestTagInfo:
        """Return the latest tag information."""
        return LatestTagInfo()

    def add_path(self, path: Pathlike) -> None:
        """Add a path to the pending commit."""
        logger.debug("No source code management system configured. Skipping adding path '%s'.", path)

    def get_all_tags(self) -> List[str]:
        """Return all tags in the SCM."""
        return []

    def commit_and_tag(self, files: List[Pathlike], context: MutableMapping, dry_run: bool = False) -> None:
        """Pretend to commit and tag files to the repository using the configuration."""
        logger.debug("No source code management system configured. Skipping committing and tagging.")

    def assert_nondirty(self) -> None:
        """Always says the repository is clean."""
        return


class SCMInfo:
    """Information about the current source code manager and state."""

    def __init__(self, config: SCMConfig):
        self.config = config
        self.commit_sha: Optional[str] = None
        self.distance_to_latest_tag: int = 0
        self.current_version: Optional[str] = None
        self.current_tag: Optional[str] = None
        self.branch_name: Optional[str] = None
        self.short_branch_name: Optional[str] = None
        self.repository_root: Optional[Path] = None
        self.dirty: Optional[bool] = None
        self.tool: SCMTool = DefaultSCMTool(config)
        self._set_from_scm_tool()

    def __repr__(self) -> str:
        """Return a string representation of the SCMInfo."""
        import pprint
        from io import StringIO

        str_io = StringIO()
        pprint.pprint(self.as_dict(), stream=str_io)
        return str_io.getvalue().rstrip()

    def as_dict(self) -> dict[str, Any]:
        """Return the information as a dict."""
        return {
            "commit_sha": self.commit_sha,
            "distance_to_latest_tag": self.distance_to_latest_tag,
            "current_version": self.current_version,
            "current_tag": self.current_tag,
            "branch_name": self.branch_name,
            "short_branch_name": self.short_branch_name,
            "repository_root": self.repository_root,
            "dirty": self.dirty,
            "tool": self.tool,
        }

    def _set_from_scm_tool(self):
        """Set information from the source code management tool."""
        from bumpversion.scm.git import Git
        from bumpversion.scm.hg import Mercurial

        git = Git(self.config)
        hg = Mercurial(self.config)
        if git.is_available():
            self.tool = git
        elif hg.is_available():
            self.tool = hg
        self._update_from_latest_tag_info(self.tool.latest_tag_info())

    def _update_from_latest_tag_info(self, latest_tag_info: LatestTagInfo):
        """Update information from the latest tag information."""
        for attr_name, value in asdict(latest_tag_info).items():
            setattr(self, attr_name, value)

    def path_in_repo(self, path: Pathlike) -> bool:
        """Return whether a path is inside this repository."""
        if self.repository_root is None:
            return True
        elif not Path(path).is_absolute():
            return True
        return str(path).startswith(str(self.repository_root))

    def commit_and_tag(self, files: List[Pathlike], context: MutableMapping, dry_run: bool = False) -> None:
        """Commit the files to the source code management system."""
        logger.indent()

        if self.config.commit:
            self._commit(files, context, dry_run)
        else:
            logger.info("Would not commit")

        if self.config.tag:
            self._tag(context, dry_run)
        else:
            logger.info("Would not tag")

        self.tool.commit_and_tag(files=files, context=context, dry_run=dry_run)
        logger.dedent()

    def _commit(self, files: List[Pathlike], context: MutableMapping, dry_run: bool = False) -> None:
        """Commit the files to the source code management system."""
        do_commit = not dry_run
        logger.info("%s the commit", "Preparing" if do_commit else "Would prepare")

        for path in files:
            logger.info("%s changes in file '%s' to repository", "Adding" if do_commit else "Would add", path)

        commit_message = self.config.message.format(**context)
        logger.info(
            "%s to repository with message '%s'", "Committing" if do_commit else "Would commit", commit_message
        )

    def _tag(self, context: MutableMapping, dry_run: bool = False) -> None:
        """Tag the current commit in the source code management system."""
        sign_tags = self.config.sign_tags
        tag_name = self.config.tag_name.format(**context)
        tag_message = self.config.tag_message.format(**context)
        existing_tags = self.tool.get_all_tags()
        do_tag = not dry_run

        if tag_name in existing_tags:
            logger.warning("Tag '%s' already exists. Will not tag.", tag_name)
            return

        logger.info(
            "%s '%s' %s to repository and %s",
            "Adding tag" if do_tag else "Would add tag",
            tag_name,
            f"with message '{tag_message}'" if tag_message else "without message",
            "signing" if sign_tags else "not signing",
        )

        for moveable_tag in self.config.moveable_tags:
            tag_name = moveable_tag.format(**context)
            logger.info("%s moveable tag '%s' in repository", "Tagging" if do_tag else "Would tag", tag_name)



"""Draw.io export script."""

import os.path
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional


def get_executable_paths() -> List[str]:
    """Get the Draw.io executable paths for the platform.

    Declared as a function to allow us to use API/environment information
    available only when running under the specified platform.

    Returns:
        All known paths.
    """
    if sys.platform.startswith("darwin"):
        applications = [os.path.expanduser("~/Applications"), "/Applications"]
        drawio_path = os.path.join("draw.io.app", "Contents", "MacOS", "draw.io")
        return [os.path.join(d, drawio_path) for d in applications]
    elif sys.platform.startswith("linux"):
        return ["/opt/draw.io/drawio"]
    elif sys.platform.startswith("win32"):
        program_files = [os.environ["ProgramFiles"]]
        if "ProgramFiles(x86)" in os.environ:
            program_files.append(os.environ["ProgramFiles(x86)"])
        return [os.path.join(d, "draw.io", "draw.io.exe") for d in program_files]
    else:
        return []


def get_drawio_executable(executable_names: Optional[List[str]] = None) -> Optional[str]:
    """Ensure the Draw.io executable path is configured, or guess it.

    Args:
        executable_names: List of executable names to check.

    Returns:
        Final Draw.io executable.
    """
    executable_names = executable_names or ["draw.io", "drawio"]

    for executable_name in executable_names:
        executable = shutil.which(executable_name)
        if executable:
            return executable

    executable_paths = get_executable_paths()

    return next(
        (executable_path for executable_path in executable_paths if os.path.isfile(executable_path)),
        None,
    )


def use_cached_file(source: Path, cache_filename: Path) -> bool:
    """Is the cached copy up to date?

    Args:
        source: Source path.
        cache_filename: Export cache filename.

    Returns:
        True if cache is up to date else False.
    """
    return cache_filename.exists() and cache_filename.stat().st_mtime >= source.stat().st_mtime


def export_file(
    source: Path,
    page_index: int,
    dest: Path,
    export_format: str,
) -> int:
    """Export an individual file.

    Args:
        source: Source path, absolute.
        page_index: Page index, numbered from zero.
        dest: Destination path, within cache.
        export_format: Export format.

    Returns:
        The Draw.io exit status.
    """
    drawio_executable = get_drawio_executable()
    cmd = [
        str(drawio_executable),
        "--export",
        str(source),
        "--page-index",
        str(page_index),
        "--output",
        str(dest),
        "--format",
        export_format,
        "--embed-svg-images",
        "--scale",
        "2",
    ]
    result = subprocess.run(cmd, check=False)  # noqa: S603
    return result.returncode


def export_file_if_needed(source: Path, page_index: int, dest_path: Path) -> None:
    """Export an individual file if needed.

    Args:
        source: Source path, absolute.
        page_index: Page index, numbered from zero.
        dest_path: Destination path.
    """
    if not use_cached_file(source, dest_path):
        export_file(source, page_index, dest_path, "svg")
    else:
        print(f"Using cached file {dest_path}")


if __name__ == "__main__":
    output_files = [
        "creating-a-version-spec.svg",
        "creating-a-version.svg",
        "serializing-a-version-1.svg",
        "serializing-a-version-1-0-0.svg",
        "serializing-a-version-1-2-0.svg",
        "serializing-a-version-1-2-3.svg",
    ]
    source_path = Path(__file__).parent.parent.joinpath("docs/assets/bump-my-version-model.drawio")
    dest_path = Path(__file__).parent.parent.joinpath("docs/assets/")
    for index, filename in enumerate(output_files):
        filepath = dest_path.joinpath(filename)
        export_file_if_needed(source_path, index, filepath)



"""Tests the bump CLI subcommand."""

import logging
import shutil
import subprocess
import traceback
from datetime import datetime, timezone
from pathlib import Path
from textwrap import dedent

import pytest
from click.testing import Result
from pytest import param

from bumpversion import cli
from tests.conftest import inside_dir


class TestNoConfiguredFilesOption:
    """Tests around the behavior of the --no-configured-files option."""

    def test_no_files_marked_to_modify(self, mocker, tmp_path: Path, runner):
        """
        No files are sent to the `do_bump` function if the --no-configured-files option is used.
        """
        mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")
        with inside_dir(tmp_path):
            result: Result = runner.invoke(
                cli.cli, ["bump", "--current-version", "1.0.0", "--no-configured-files", "patch"]
            )

        if result.exit_code != 0:
            print(result.output)

        assert result.exit_code == 0

        call_args = mocked_do_bump.call_args[0]
        passed_config = call_args[2]
        assert len(passed_config.files) == 0

    def test_file_args_marked_to_modify(self, mocker, tmp_path: Path, runner):
        """File paths marked in the command-line arguments are sent to the `do_bump` function."""
        mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")
        with inside_dir(tmp_path):
            result: Result = runner.invoke(
                cli.cli,
                ["bump", "--current-version", "1.0.0", "--no-configured-files", "patch", "do-this-file.txt"],
            )

        if result.exit_code != 0:
            print(result.output)

        assert result.exit_code == 0

        call_args = mocked_do_bump.call_args[0]
        passed_config = call_args[2]
        assert len(passed_config.files) == 1
        assert passed_config.files[0].filename == "do-this-file.txt"


def test_bump_nested_regex(tmp_path: Path, fixtures_path: Path, caplog, runner):
    """
    Arrange/Act: Run the `bump` subcommand with --no-configured-files.

    Assert: There is no configured files specified to modify
    """
    cff_path = tmp_path / "citation.cff"
    cff_path.write_text("cff-version: 1.2.0\ndate-released: 2023-09-19\n", encoding="utf-8")
    content = fixtures_path.joinpath("regex_test_config.toml").read_text()
    config_path = tmp_path / ".bumpversion.toml"
    config_path.write_text(content, encoding="utf-8")

    with inside_dir(tmp_path):
        result: Result = runner.invoke(cli.cli, ["bump", "-vv", "patch"])

    if result.exit_code != 0:
        print(result.output)
        print(caplog.text)

    assert result.exit_code == 0

    now = datetime.now(timezone.utc).isoformat()[:10]
    assert cff_path.read_text() == f"cff-version: 1.2.0\ndate-released: {now}\n"


def test_missing_explicit_config_file(tmp_path: Path, runner):
    """The command-line processor should raise an exception if the config file is missing."""
    with inside_dir(tmp_path):
        with inside_dir(tmp_path):
            result: Result = runner.invoke(cli.cli, ["bump", "--config-file", "missing-file.cfg"])
        assert result.exit_code != 0
        assert "'missing-file.cfg' does not exist." in result.output


def test_cli_options_override_config(tmp_path: Path, fixtures_path: Path, mocker, runner):
    """The command-line processor should override the config file."""
    # Arrange
    config_path = tmp_path / "this_config.toml"
    fixture_toml = fixtures_path / "basic_cfg.toml"
    shutil.copy(fixture_toml, config_path)
    mocked_do_bump = mocker.patch("bumpversion.cli.do_bump")

    # Act
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "--config-file",
                str(config_path),
                "--current-version",
                "1.1.0",
                "--new-version",
                "1.2.0",
                "--allow-dirty",
                "--parse",
                r"XXX(?P<spam>\d+);(?P<blob>\d+);(?P<slurp>\d+)",
                "--serialize",
                "XXX{spam};{blob};{slurp}",
                "--search",
                "my-search",
                "--replace",
                "my-replace",
                "--no-commit",
                "--no-tag",
                "slurp",
                "do-this-file.txt",
            ],
        )

    # Assert
    assert result.exit_code == 0
    assert mocked_do_bump.call_count == 1
    assert mocked_do_bump.call_args[0][0] == "slurp"
    assert mocked_do_bump.call_args[0][1] == "1.2.0"
    the_config = mocked_do_bump.call_args[0][2]
    assert the_config.current_version == "1.1.0"
    assert the_config.allow_dirty
    assert the_config.parse == r"XXX(?P<spam>\d+);(?P<blob>\d+);(?P<slurp>\d+)"
    assert the_config.serialize == ("XXX{spam};{blob};{slurp}",)
    assert the_config.search == "my-search"
    assert the_config.replace == "my-replace"
    assert the_config.commit is False
    assert the_config.tag is False
    assert len(the_config.files) == 4
    assert {f.filename for f in the_config.files} == {
        "setup.py",
        "bumpversion/__init__.py",
        "CHANGELOG.md",
        "do-this-file.txt",
    }
    assert mocked_do_bump.call_args[0][3] == config_path


@pytest.mark.parametrize(
    ["repo", "scm_command"],
    [
        param(
            "git_repo",
            "git",
            id="git",
            marks=pytest.mark.skipif(not shutil.which("git"), reason="Git is not available."),
        ),
        param(
            "hg_repo",
            "hg",
            id="hg",
            marks=pytest.mark.skipif(not shutil.which("hg"), reason="Mercurial is not available."),
        ),
    ],
)
def test_dirty_work_dir_raises_error(repo: str, scm_command: str, request, runner):
    """A dirty working directory should raise an error."""
    repo_path: Path = request.getfixturevalue(repo)
    with inside_dir(repo_path):
        # Arrange
        repo_path.joinpath("dirty2").write_text("i'm dirty! 1.1.1", encoding="utf-8")
        subprocess.run([scm_command, "add", "dirty2"], check=True)

        # Act
        result: Result = runner.invoke(
            cli.cli, ["bump", "patch", "--current-version", "1.1.1", "-vv", "--no-allow-dirty", "dirty2"]
        )

    # Assert
    assert "working directory is not clean" in result.output


def test_non_scm_operations_if_scm_not_installed(tmp_path: Path, monkeypatch, runner):
    """Everything works except SCM commands if the SCM is unusable."""
    # Arrange
    monkeypatch.setenv("PATH", "")

    with inside_dir(tmp_path):
        version_path = tmp_path / "VERSION"
        version_path.write_text("31.0.3", encoding="utf-8")

        # Act
        runner.invoke(cli.cli, ["bump", "major", "--current-version", "31.0.3", "VERSION"])

    # Assert
    assert version_path.read_text() == "32.0.0"


@pytest.mark.parametrize(
    ["version_part"],
    [
        param("charlie", id="bad_version_component"),
        param("", id="missing_version_component"),
    ],
)
def test_detects_bad_or_missing_version_component(version_part: str, tmp_path: Path, monkeypatch, runner):
    """It properly detects bad or missing version part."""
    # Arrange
    monkeypatch.setenv("PATH", "")

    with inside_dir(tmp_path):
        version_path = tmp_path / "VERSION"
        version_path.write_text("31.0.3", encoding="utf-8")

        args = ["bump", "--current-version", "31.0.3"]
        if version_part:
            args.append(version_part)
        args.append("VERSION")
        # Act
        result = runner.invoke(cli.cli, args)

    # Assert
    assert result.exception is not None
    assert "Unknown version component:" in result.stdout


def test_ignores_missing_files_with_option(tmp_path, fixtures_path, runner):
    """The replace subcommand should ignore missing."""

    config_file = tmp_path / ".bumpversion.toml"
    config_file.write_text(
        "[tool.bumpversion]\n"
        'current_version = "0.0.1"\n'
        "allow_dirty = true\n\n"
        "[[tool.bumpversion.files]]\n"
        'filename = "VERSION"\n'
        "regex = false\n",
        encoding="utf-8",
    )

    # Act
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "--verbose",
                "--no-regex",
                "--no-configured-files",
                "--ignore-missing-files",
                "minor",
                "VERSION",
            ],
        )

    # Assert
    if result.exit_code != 0:
        print("Here is the output:")
        print(result.output)
        print(traceback.print_exception(result.exc_info[1]))

    assert result.exit_code == 0


def test_dash_in_tag_matches_current_version(git_repo, runner, caplog):
    import logging

    caplog.set_level(logging.DEBUG, logger="bumpversion")

    repo_path: Path = git_repo
    with inside_dir(repo_path):
        # Arrange
        config_toml = dedent(
            """
            [tool.bumpversion]
            current_version = "4.0.3-beta+build.1047"
            parse = "(?P<major>\\\\d+)\\\\.(?P<minor>\\\\d+)\\\\.(?P<patch>\\\\d+)(-(?P<release>[0-9A-Za-z]+))?(\\\\+build\\\\.(?P<build>.[0-9A-Za-z]+))?"
            serialize = [
                "{major}.{minor}.{patch}-{release}+build.{build}",
                "{major}.{minor}.{patch}+build.{build}"
            ]
            commit = true
            message = "Bump version: {current_version} -> {new_version}"
            tag = false
            tag_name = "app/{new_version}"
            tag_message = "Version app/{new_version}"
            allow_dirty = false

            [tool.bumpversion.parts.release]
            values = [
                "beta",
                "rc",
                "final"
            ]
            optional_value = "final"

            [tool.bumpversion.parts.build]
            first_value = 1000
            independent = true
            always_increment = true
        """
        )
        repo_path.joinpath(".bumpversion.toml").write_text(config_toml, encoding="utf-8")
        subprocess.run(["git", "add", ".bumpversion.toml"], check=True)
        subprocess.run(["git", "commit", "-m", "added file"], check=True)
        subprocess.run(["git", "tag", "-a", "app/4.0.3-beta+build.1047", "-m", "added tag"], check=True)

        # Act
        result: Result = runner.invoke(cli.cli, ["bump", "-vv", "--tag", "--commit", "patch"])

    # Assert
    if result.exit_code != 0:
        import traceback

        print(caplog.text)
        print(traceback.print_exception(result.exc_info[1]))

    assert result.exit_code == 0


def test_current_version_not_required_in_config(tmp_path, fixtures_path, runner):
    """The current version is not required in the configuration."""

    config_file = tmp_path / ".bumpversion.toml"
    config_contents = "[tool.bumpversion]\nallow_dirty = true\n\n"
    config_file.write_text(
        config_contents,
        encoding="utf-8",
    )

    # Act
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "-vv",
                "--current-version",
                "1.0.0",
                "--dry-run",
                "minor",
            ],
        )

    # Assert
    if result.exit_code != 0:
        print("Here is the output:")
        print(result.output)
        print(traceback.print_exception(result.exc_info[1]))

    assert result.exit_code == 0
    assert config_file.read_text() == config_contents


def test_remote_config_doesnt_update(tmp_path, git_repo, runner):
    """The current version is not required in the configuration."""

    config_file = tmp_path / ".bumpversion.toml"
    config_contents = '[tool.bumpversion]\nallow_dirty = true\n\n[[tool.bumpversion.files]]\nfilename = "readme.md"\n'
    config_file.write_text(
        config_contents,
        encoding="utf-8",
    )
    myfile = git_repo / "readme.md"
    myfile.write_text("version = 1.0.0\n", encoding="utf-8")

    # Act
    with inside_dir(git_repo):
        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "-vv",
                "--current-version",
                "1.0.0",
                f"--config-file={config_file}",
                "minor",
            ],
        )

    # Assert
    if result.exit_code != 0:
        print("Here is the output:")
        print(result.output)
        print(traceback.print_exception(result.exc_info[1]))

    assert result.exit_code == 0
    assert config_file.read_text() == config_contents
    assert myfile.read_text() == "version = 1.1.0\n"


def test_moveable_tags(git_repo: Path, fixtures_path: Path, runner, caplog):
    """The current version is not required in the configuration."""
    caplog.set_level(logging.DEBUG)
    config_file = git_repo / ".bumpversion.toml"
    orig_config = fixtures_path / "basic_cfg_moveable.toml"
    shutil.copy(orig_config, config_file)
    changelog_path = git_repo / "CHANGELOG.md"
    changelog_path.write_text("**unreleased**")
    mkdir_path = git_repo / "example"
    mkdir_path.mkdir()
    init_path = mkdir_path / "__init__.py"
    init_path.write_text('__version__ = "1.0.0"')

    # Act
    with inside_dir(git_repo):
        subprocess.run(["git", "add", "."], check=True)
        subprocess.run(["git", "commit", "-m", "added files"], check=True)

        result: Result = runner.invoke(
            cli.cli,
            [
                "bump",
                "-vv",
                "--dry-run",
                "minor",
            ],
        )

    # Assert
    if result.exit_code != 0:
        print("Here is the output:")
        print(result.output)
        print(traceback.print_exception(*result.exc_info))

    assert result.exit_code == 0
    assert "Would tag moveable tag 'v1'" in caplog.text



"""Functions for displaying information about the version."""

import dataclasses
from io import StringIO
from pathlib import Path
from pprint import pprint
from typing import Any, Optional, Union

from bumpversion.bump import get_next_version
from bumpversion.config import Config
from bumpversion.context import get_context
from bumpversion.exceptions import BadInputError
from bumpversion.scm.models import SCMInfo
from bumpversion.ui import print_error, print_info
from bumpversion.utils import recursive_sort_dict


def output_default(value: dict) -> None:
    """Output the value with key=value or just value if there is only one item."""
    if len(value) == 1:
        print_info(next(iter(value.values())))
    else:
        buffer = StringIO()
        pprint(value, stream=buffer, sort_dicts=True)
        print_info(buffer.getvalue())


def output_yaml(value: dict) -> None:
    """Output the value as yaml."""
    from bumpversion.yaml_dump import dump

    print_info(dump(recursive_sort_dict(value)))


def output_json(value: dict) -> None:
    """Output the value as json."""
    import json

    def default_encoder(obj: Any) -> Union[str, dict]:
        if dataclasses.is_dataclass(obj):
            return str(obj)
        elif isinstance(obj, type):
            return obj.__name__
        elif isinstance(obj, Path):
            return str(obj)
        elif isinstance(obj, SCMInfo):
            return obj.as_dict()
        return str(obj)

    print_info(json.dumps(value, sort_keys=True, indent=2, default=default_encoder))


OUTPUTTERS = {
    "yaml": output_yaml,
    "json": output_json,
    "default": output_default,
}


def resolve_name(obj: Any, name: str, default: Any = None, err_on_missing: bool = False) -> Any:
    """
    Get a key or attr ``name`` from obj or default value.

    Copied and modified from Django Template variable resolutions

    Resolution methods:

    - Mapping key lookup
    - Attribute lookup
    - Sequence index

    Args:
        obj: The object to access
        name: A dotted name to the value, such as ``mykey.0.name``
        default: If the name cannot be resolved from the object, return this value
        err_on_missing: Raise a `BadInputError` if the name cannot be resolved

    Returns:
        The value at the resolved name or the default value.

    Raises:
        BadInputError: If we cannot resolve the name and `err_on_missing` is `True`
        AttributeError: If a @property decorator raised it
        TypeError: If a @property decorator raised it

    # noqa: DAR401
    """
    lookups = name.split(".")
    current = obj
    try:  # catch-all for unexpected failures
        for bit in lookups:
            try:  # dictionary lookup
                current = current[bit]
                # ValueError/IndexError are for numpy.array lookup on
                # numpy < 1.9 and 1.9+ respectively
            except (TypeError, AttributeError, KeyError, ValueError, IndexError):
                try:  # attribute lookup
                    current = getattr(current, bit)
                except (TypeError, AttributeError):
                    # Reraise if the exception was raised by a @property
                    if bit in dir(current):
                        raise
                    try:  # list-index lookup
                        current = current[int(bit)]
                    except (
                        IndexError,  # list index out of range
                        ValueError,  # invalid literal for int()
                        KeyError,  # current is a dict without `int(bit)` key
                        TypeError,
                    ):  # un-subscript-able object
                        return default
        return current
    except Exception as e:  # pragma: no cover
        if err_on_missing:
            raise BadInputError(f"Could not resolve '{name}'") from e
        else:
            return default


def do_show(
    *args,
    config: Config,
    format_: str = "default",
    increment: Optional[str] = None,
    current_version: Optional[str] = None,
) -> None:
    """Show current version or configuration information."""
    if current_version:
        config.current_version = current_version
    config_dict = config.model_dump()
    ctx = get_context(config)

    if increment:
        version = config.version_config.parse(config.current_version)
        next_version = get_next_version(version, config, increment, None)
        next_version_str = config.version_config.serialize(next_version, ctx)
        config_dict["new_version"] = next_version_str

    try:
        if "all" in args or not args:
            show_items = config_dict
        else:
            show_items = {key: resolve_name(config_dict, key) for key in args}

        OUTPUTTERS.get(format_, OUTPUTTERS["default"])(show_items)
    except BadInputError as e:
        print_error(e.message)



"""Mercurial source control management."""

import json
import os
import re
import shlex
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import ClassVar, List, MutableMapping, Optional

from bumpversion.exceptions import DirtyWorkingDirectoryError, SignedTagsError
from bumpversion.scm.models import LatestTagInfo, SCMConfig
from bumpversion.ui import get_indented_logger
from bumpversion.utils import Pathlike, format_and_raise_error, is_subpath, run_command

logger = get_indented_logger(__name__)


class Mercurial:
    """Mercurial source control management."""

    _TEST_AVAILABLE_COMMAND: ClassVar[List[str]] = ["hg", "root"]
    _COMMIT_COMMAND: ClassVar[List[str]] = ["hg", "commit", "--logfile"]

    def __init__(self, config: SCMConfig):
        self.config = config
        self._latest_tag_info: Optional[LatestTagInfo] = None

    def __repr__(self) -> str:
        """Return a string representation of the SCMTool."""
        return self.__str__()

    def __str__(self) -> str:
        """A string representation of the object."""
        return "Mercurial"

    def is_available(self) -> bool:
        """Is the VCS implementation usable?"""
        try:
            result = run_command(self._TEST_AVAILABLE_COMMAND)
            return result.returncode == 0
        except (FileNotFoundError, PermissionError, NotADirectoryError, subprocess.CalledProcessError):
            return False

    def latest_tag_info(self) -> LatestTagInfo:
        """Return information about the latest tag."""
        if self._latest_tag_info is not None:
            return self._latest_tag_info

        if not self.is_available():
            return LatestTagInfo()

        info = commit_info(self.config)
        self._latest_tag_info = LatestTagInfo(**info)
        return self._latest_tag_info

    def get_all_tags(self) -> List[str]:
        """Return all tags in a mercurial repository."""
        try:
            result = run_command(["hg", "tags", "-T", "json"])
            tags = json.loads(result.stdout) if result.stdout else []
            return [tag["tag"] for tag in tags]
        except (
            FileNotFoundError,
            PermissionError,
            NotADirectoryError,
            subprocess.CalledProcessError,
        ) as e:
            format_and_raise_error(e)
            return []

    def add_path(self, path: Pathlike) -> None:
        """Add a path to the Source Control Management repository."""
        repository_root = self.latest_tag_info().repository_root
        if not (repository_root and is_subpath(repository_root, path)):
            return

        cwd = Path.cwd()
        temp_path = os.path.relpath(path, cwd)
        try:
            run_command(["hg", "add", str(temp_path)])
        except subprocess.CalledProcessError as e:  # pragma: no-cover
            format_and_raise_error(e)

    def commit_and_tag(self, files: List[Pathlike], context: MutableMapping, dry_run: bool = False) -> None:
        """Commit and tag files to the repository using the configuration."""
        if dry_run:
            return

        if self.config.commit:
            for path in files:
                self.add_path(path)

            self.commit(context)

        if self.config.tag:
            tag_name = self.config.tag_name.format(**context)
            tag_message = self.config.tag_message.format(**context)
            tag(tag_name, sign=self.config.sign_tags, message=tag_message)

            # for m_tag_name in self.config.moveable_tags:
            #     moveable_tag(m_tag_name)

    def commit(self, context: MutableMapping) -> None:
        """Commit the changes."""
        extra_args = shlex.split(self.config.commit_args) if self.config.commit_args else []

        current_version = context.get("current_version", "")
        new_version = context.get("new_version", "")
        commit_message = self.config.message.format(**context)

        if not current_version:  # pragma: no-coverage
            logger.warning("No current version given, using an empty string.")
        if not new_version:  # pragma: no-coverage
            logger.warning("No new version given, using an empty string.")

        with NamedTemporaryFile("wb", delete=False) as f:
            f.write(commit_message.encode("utf-8"))

        env = os.environ.copy()
        env["BUMPVERSION_CURRENT_VERSION"] = current_version
        env["BUMPVERSION_NEW_VERSION"] = new_version

        try:
            cmd = [*self._COMMIT_COMMAND, f.name, *extra_args]
            run_command(cmd, env=env)
        except (subprocess.CalledProcessError, TypeError) as exc:  # pragma: no-coverage
            format_and_raise_error(exc)
        finally:
            os.unlink(f.name)

    @staticmethod
    def tag(name: str, sign: bool = False, message: Optional[str] = None) -> None:
        """
        Create a tag of the new_version in VCS.

        If only name is given, bumpversion uses a lightweight tag.
        Otherwise, it uses an annotated tag.

        Args:
            name: The name of the tag
            sign: True to sign the tag
            message: A optional message to annotate the tag.

        Raises:
            SignedTagsError: If ``sign`` is ``True``
        """
        tag(name, sign=sign, message=message)

    @staticmethod
    def assert_nondirty() -> None:
        """Assert that the working directory is clean."""
        assert_nondirty()


def commit_info(config: SCMConfig) -> dict:
    """
    Get the commit info for the repo.

    Args:
        config: The source control configuration.

    Returns:
        A dictionary containing information about the latest commit.
    """
    tag_pattern = config.tag_name.replace("{new_version}", ".*")
    info = dict.fromkeys(
        [
            "dirty",
            "commit_sha",
            "distance_to_latest_tag",
            "current_version",
            "current_tag",
            "branch_name",
            "short_branch_name",
            "repository_root",
        ]
    )

    info["distance_to_latest_tag"] = 0
    result = run_command(["hg", "log", "-r", f"tag('re:{tag_pattern}')", "-T", "json"])
    repo_path = run_command(["hg", "root"]).stdout.strip()

    output_info = parse_commit_log(result.stdout, config)
    info |= output_info

    if not output_info:
        logger.debug("No tags found")

    info["repository_root"] = Path(repo_path)
    info["dirty"] = len(run_command(["hg", "status", "-mard"]).stdout) != 0
    return info


def parse_commit_log(log_string: str, config: SCMConfig) -> dict:
    """Parse the commit log string."""
    output_info = json.loads(log_string) if log_string else {}
    if not output_info:
        return {}
    first_rev = output_info[0]
    branch_name = first_rev["branch"]
    short_branch_name = re.sub(r"([^a-zA-Z0-9]*)", "", branch_name).lower()[:20]

    return {
        "current_version": config.get_version_from_tag(first_rev["tags"][0]),
        "current_tag": first_rev["tags"][0],
        "commit_sha": first_rev["node"],
        "distance_to_latest_tag": 0,
        "branch_name": branch_name,
        "short_branch_name": short_branch_name,
    }


def tag(name: str, sign: bool = False, message: Optional[str] = None) -> None:
    """
    Create a tag of the new_version in VCS.

    If only name is given, bumpversion uses a lightweight tag.
    Otherwise, it uses an annotated tag.

    Args:
        name: The name of the tag
        sign: True to sign the tag
        message: A optional message to annotate the tag.

    Raises:
        SignedTagsError: If ``sign`` is ``True``
    """
    command = ["hg", "tag", name]
    if sign:
        raise SignedTagsError("Mercurial does not support signed tags.")
    if message:
        command += ["--message", message]
    run_command(command)


def assert_nondirty() -> None:
    """Assert that the working directory is clean."""
    if lines := [
        line.strip()
        for line in run_command(["hg", "status", "-mard"]).stdout.splitlines()
        if not line.strip().startswith("??")
    ]:
        joined_lines = "\n".join(lines)
        raise DirtyWorkingDirectoryError(f"Mercurial working directory is not clean:\n{joined_lines}")



#!/usr/bin/env python
"""Update frontmatter of markdown files."""

import argparse
from pathlib import Path
from typing import Any, Dict, Optional

import frontmatter


def extract_main_heading(markdown_content: str) -> Optional[str]:
    """
    Extracts the first level 1 heading from the provided Markdown content.

    Args:
        markdown_content: A string containing Markdown text.

    Returns:
        The text of the first level 1 heading, or None if no such heading is found.
    """
    lines = markdown_content.split("\n")

    return next((line[2:] for line in lines if line.startswith("# ")), None)


def calc_title(post: frontmatter.Post) -> str:
    """Calculate the title of the post."""
    return extract_main_heading(post.content) or post.get("title", "")


def calc_comment(post: frontmatter.Post) -> bool:
    """Calculate if the post has comments."""
    return bool(post.get("comments", True))


def calculate_update(post: frontmatter.Post) -> dict:
    """Calculate if the frontmatter needs to be updated."""
    expected_title = calc_title(post)
    expected_comment = calc_comment(post)
    update: Dict[str, Any] = {}
    if expected_title and expected_title != post.get("title"):
        update["title"] = expected_title
    if expected_comment != post.get("comments"):
        update["comments"] = expected_comment
    return update


def process_file(markdown_path: Path) -> None:
    """Process a single file."""
    if not (markdown_path.is_file() and markdown_path.suffix == ".md"):
        return
    raw_text = markdown_path.read_text()
    post = frontmatter.loads(raw_text)

    update = calculate_update(post)
    if update:
        for key, value in update.items():
            post[key] = value
        new_text = frontmatter.dumps(post)
        print(f"Updating {markdown_path}")
        markdown_path.write_text(new_text, encoding="utf-8")


def parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Update frontmatter of markdown files")
    parser.add_argument("markdown_path", type=str, nargs="+", help="Path or glob to markdown files")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    documents = args.markdown_path
    for document in documents:
        for path in Path().glob(document):
            process_file(path)



"""Tests for bumpversion.scm.models."""

import pytest
from pytest import param

from bumpversion.scm.models import SCMConfig

simple_pattern = r"(?P<major>\\d+)\\.(?P<minor>\\d+)\.(?P<patch>\\d+)"
complex_pattern = (
    r"(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)"
    r"(-(?P<release>[0-9A-Za-z]+))?(\\+build\\.(?P<build>.[0-9A-Za-z]+))?"
)


@pytest.mark.parametrize(
    ["tag", "tag_name", "parse_pattern", "expected"],
    [
        param("1.2.3", "{new_version}", simple_pattern, "1.2.3", id="no-prefix, no-suffix"),
        param("v/1.2.3", "v/{new_version}", simple_pattern, "1.2.3", id="prefix, no-suffix"),
        param("v/1.2.3/12345", "v/{new_version}/{something}", simple_pattern, "1.2.3", id="prefix, suffix"),
        param("1.2.3/2024-01-01", "{new_version}/{today}", simple_pattern, "1.2.3", id="no-prefix, suffix"),
        param(
            "app/4.0.3-beta+build.1047",
            "app/{new_version}",
            complex_pattern,
            "4.0.3-beta+build.1047",
            id="complex",
        ),
    ],
)
def test_get_version_from_tag_returns_version_from_pattern(
    scm_config: SCMConfig, tag: str, tag_name: str, parse_pattern: str, expected: str
) -> None:
    """SCMConfig.get_version_from_tag properly returns the version from the tag."""
    # Arrange
    scm_config.parse_pattern = parse_pattern
    scm_config.tag_name = tag_name

    # Act
    version = scm_config.get_version_from_tag(tag)

    # Assert
    assert version == expected
